from __future__ import annotations

from typing import Any, Dict, List, Union

from boa3 import constants
from boa3.neo.utils import contract_parameter_to_json, stack_item_from_json
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3.neo3.core.serialization import BinaryReader
from boa3_test.tests.test_classes.nativecontractprefix import get_native_contract_data


class Storage:
    def __init__(self):
        self._dict: Dict[StorageKey, StorageItem] = {}

    def pop(self, key: bytes) -> StorageItem:
        storage_key = StorageKey(key)
        return self._dict.pop(storage_key)

    def clear(self, delete_deploy_data: bool = True):
        prefix, native_id = get_native_contract_data(constants.MANAGEMENT_SCRIPT)
        for key in list(self._dict.keys()):
            should_delete = (key._ID > 0  # it's a deployd contract
                             or (delete_deploy_data
                                 and key._ID == native_id
                                 and key._key.startswith(prefix)  # it's a deployed contract register
                                 )
                             )
            if should_delete:
                # keep native contracts storage
                self._dict.pop(key)

    def copy(self) -> Storage:
        storage = Storage()
        storage._dict = self._dict.copy()
        return storage

    def to_json(self) -> List[Dict[str, Any]]:
        return [{'key': key.to_json(),
                 'value': item.to_json()
                 } for key, item in self._dict.items()
                ]

    @classmethod
    def from_json(cls, json: Union[Dict[str, Any], List[Dict[str, Any]]]) -> Storage:
        if not isinstance(json, list):
            json = [json]

        new_storage = {}
        for storage_value in json:
            key = StorageKey.from_json(storage_value['key'])
            value = StorageItem.from_json(storage_value['value'])
            new_storage[key] = value

        storage = Storage()
        storage._dict = new_storage
        return storage

    def add_token(self, token_script: bytes, script_hash: bytes, amount: int) -> bool:
        if len(token_script) != 20 or len(script_hash) != 20 or amount <= 0:
            return False

        from boa3_test.tests.test_classes.nativecontractprefix import get_native_contract_data
        token_prefix, token_id = get_native_contract_data(token_script)
        if token_prefix is None or token_id is None:
            return False

        balance_key = token_prefix + script_hash
        if balance_key in self._dict:
            balance = Integer.from_bytes(self[balance_key])
        else:
            balance = 0

        balance += amount

        from boa3_test.tests.test_classes.nativeaccountstate import NativeAccountState
        key = StorageKey(balance_key)
        key._ID = token_id
        self._dict[key] = StorageItem(NativeAccountState(balance).serialize())
        return True

    def has_contract(self, script_hash: bytes) -> bool:
        prefix, native_id = get_native_contract_data(constants.MANAGEMENT_SCRIPT)
        if prefix is None or native_id is None:
            return False

        storage_key = self.build_key(prefix + script_hash, native_id)
        return storage_key in self

    def get_contract_id(self, script_hash: bytes) -> int:
        prefix, native_id = get_native_contract_data(constants.MANAGEMENT_SCRIPT)
        if prefix is None or native_id is None:
            return False

        storage_key = self.build_key(prefix + script_hash, native_id)
        if storage_key in self:
            result = self[storage_key]
            with BinaryReader(result) as reader:
                from boa3_test.tests.test_classes.binaryserializer import deserialize_binary
                result = deserialize_binary(reader)

            return result[0] if isinstance(result[0], int) else -1
        else:
            return -1

    def __contains__(self, item: StorageKey) -> bool:
        return item in self._dict

    def __getitem__(self, item: StorageKey) -> Any:
        return self._dict[item].value

    def __setitem__(self, key: StorageKey, value: Any):
        from boa3.neo.vm.type import StackItem
        if isinstance(value, int):
            storage_value = Integer(value).to_byte_array()
        elif isinstance(value, str):
            storage_value = String(value).to_bytes()
        else:
            storage_value = StackItem.serialize(value)
        self._dict[key] = StorageItem(storage_value)

    @staticmethod
    def build_key(key: bytes, index: int) -> StorageKey:
        return StorageKey(key, index)


class StorageKey:
    def __init__(self, key: bytes, _id: int = 0):
        self._ID: int = _id
        self._key: bytes = key

    def to_json(self) -> Dict[str, Any]:
        return {'id': self._ID,
                'key': contract_parameter_to_json(self._key)
                }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> StorageKey:
        k = stack_item_from_json(json['key'])
        if isinstance(k, str):
            from boa3.neo.vm.type.String import String
            k = String(k).to_bytes()

        key = StorageKey(k)
        key._ID = json['id']
        return key

    def __eq__(self, other) -> bool:
        return isinstance(other, StorageKey) and self._key == other._key and self._ID == other._ID

    def __str__(self) -> str:
        return '({0}, {1})'.format(self._key, self._ID)

    def __hash__(self) -> int:
        return self._key.__hash__()


class StorageItem:
    def __init__(self, value: bytes, is_constant: bool = False):
        self._is_constant: bool = is_constant
        self._value: bytes = value

    @property
    def value(self) -> bytes:
        return self._value

    def to_json(self) -> Dict[str, Any]:
        return {'isconstant': self._is_constant,
                'value': contract_parameter_to_json(self._value)
                }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> StorageItem:
        value = stack_item_from_json(json['value'])
        if isinstance(value, str):
            from boa3.neo.vm.type.String import String
            value = String(value).to_bytes()

        item = StorageItem(value, json['isconstant'])
        return item

    def __str__(self) -> str:
        return self._value.__str__()



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.builtin.interop.interop import Interop
from boa3.model.type.type import Type
from boa3.neo.core.types.InteropInterface import InteropInterface
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestStorageInterop(BoaTest):
    default_folder: str = 'test_sc/interop_test/storage'

    def test_storage_get_bytes_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StorageGet.interop_method_hash
            + Opcode.DUP
            + Opcode.ISNULL
            + Opcode.JMPIFNOT
            + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.PUSHDATA1
            + Integer(0).to_byte_array(signed=False, min_length=1)
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = self.get_contract_path('StorageGetBytesKey.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

    def test_storage_get_str_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StorageGet.interop_method_hash
            + Opcode.DUP
            + Opcode.ISNULL
            + Opcode.JMPIFNOT
            + Integer(7).to_byte_array(signed=True, min_length=1)
            + Opcode.DROP
            + Opcode.PUSHDATA1
            + Integer(0).to_byte_array(signed=False, min_length=1)
            + Opcode.CONVERT
            + Type.bytes.stack_item
            + Opcode.RET
        )

        path = self.get_contract_path('StorageGetStrKey.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'example',
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)

        storage = {('example', path): 23}
        result = self.run_smart_contract(engine, path, 'Main', 'example',
                                         fake_storage=storage,
                                         expected_result_type=bytes)
        self.assertEqual(Integer(23).to_byte_array(), result)

        storage = {('test1', path): 23, ('test2', path): 42}
        result = self.run_smart_contract(engine, path, 'Main', 'test2',
                                         fake_storage=storage,
                                         expected_result_type=bytes)
        self.assertEqual(Integer(42).to_byte_array(), result)

    def test_storage_get_mismatched_type(self):
        path = self.get_contract_path('StorageGetMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_storage_put_bytes_key_bytes_value(self):
        path = self.get_contract_path('StoragePutBytesKeyBytesValue.py')

        engine = TestEngine()
        stored_value = b'\x01\x02\x03'
        result = self.run_smart_contract(engine, path, 'Main', b'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_put_bytes_key_int_value(self):
        value = Integer(123).to_byte_array()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHINT8 + value
            + Opcode.STLOC0
            + Opcode.PUSHINT8 + value
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StoragePutBytesKeyIntValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        stored_value = Integer(123).to_byte_array()
        result = self.run_smart_contract(engine, path, 'Main', b'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_put_bytes_key_str_value(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.STLOC0
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StoragePutBytesKeyStrValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        stored_value = String('123').to_bytes()
        result = self.run_smart_contract(engine, path, 'Main', 'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', 'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', 'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_put_str_key_bytes_value(self):
        path = self.get_contract_path('StoragePutStrKeyBytesValue.py')

        engine = TestEngine()
        stored_value = b'\x01\x02\x03'
        result = self.run_smart_contract(engine, path, 'Main', b'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_put_str_key_int_value(self):
        value = Integer(123).to_byte_array()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHINT8 + value
            + Opcode.STLOC0
            + Opcode.PUSHINT8 + value
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StoragePutStrKeyIntValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        stored_value = Integer(123).to_byte_array()
        result = self.run_smart_contract(engine, path, 'Main', 'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', 'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', 'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_put_str_key_str_value(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.STLOC0
            + Opcode.PUSHDATA1
            + Integer(len(value)).to_byte_array(min_length=1, signed=True)
            + value
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StoragePut.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StoragePutStrKeyStrValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        stored_value = String('123').to_bytes()
        result = self.run_smart_contract(engine, path, 'Main', 'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', 'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', 'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_put_mismatched_type_key(self):
        path = self.get_contract_path('StoragePutMismatchedTypeKey.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_storage_put_mismatched_type_value(self):
        path = self.get_contract_path('StoragePutMismatchedTypeValue.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_storage_delete_bytes_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StorageDelete.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StorageDeleteBytesKey.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'example')
        self.assertIsVoid(result)
        self.assertIsNone(engine.storage_get(b'example', path))

        storage = {('example', path): 23}
        result = self.run_smart_contract(engine, path, 'Main', b'example', fake_storage=storage)
        self.assertIsVoid(result)
        self.assertIsNone(engine.storage_get('example', path))
        self.assertIsNone(engine.storage_get(b'example', path))

    def test_storage_delete_str_key(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.SYSCALL
            + Interop.StorageDelete.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StorageDeleteStrKey.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'example')
        self.assertIsVoid(result)
        self.assertIsNone(engine.storage_get(b'example', path))

        storage = {('example', path): 23}
        result = self.run_smart_contract(engine, path, 'Main', 'example', fake_storage=storage)
        self.assertIsVoid(result)
        self.assertIsNone(engine.storage_get('example', path))
        self.assertIsNone(engine.storage_get(b'example', path))

    def test_storage_delete_mismatched_type(self):
        path = self.get_contract_path('StorageDeleteMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_storage_find_bytes_prefix(self):
        path = self.get_contract_path('StorageFindBytesPrefix.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'find_by_prefix', b'example')
        self.assertEqual([], result)

        storage = {('example_0', path): '0',
                   ('example_1', path): '1',
                   ('example_2', path): '3'}
        expected_result = [[key, value] for (key, sc), value in storage.items()]

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'example', fake_storage=storage)
        self.assertEqual(expected_result, result)

    def test_storage_find_str_prefix(self):
        path = self.get_contract_path('StorageFindStrPrefix.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'example')
        self.assertEqual([], result)

        storage = {('example_0', path): '0',
                   ('example_1', path): '1',
                   ('example_2', path): '3'}
        expected_result = [[key, value] for (key, sc), value in storage.items()]

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'example', fake_storage=storage)
        self.assertEqual(expected_result, result)

    def test_storage_find_mismatched_type(self):
        path = self.get_contract_path('StorageFindMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_storage_get_context(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.RET
        )
        path = self.get_contract_path('StorageGetContext.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(InteropInterface, result)  # returns an interop interface

    def test_storage_get_read_only_context(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.StorageGetReadOnlyContext.interop_method_hash
            + Opcode.RET
        )
        path = self.get_contract_path('StorageGetReadOnlyContext.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(InteropInterface, result)  # returns an interop interface

    def test_storage_get_with_context(self):
        path = self.get_contract_path('StorageGetWithContext.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'example',
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)

        storage = {('example', path): 23}
        result = self.run_smart_contract(engine, path, 'Main', 'example',
                                         fake_storage=storage,
                                         expected_result_type=bytes)
        self.assertEqual(Integer(23).to_byte_array(), result)

        storage = {('test1', path): 23, ('test2', path): 42}
        result = self.run_smart_contract(engine, path, 'Main', 'test2',
                                         fake_storage=storage,
                                         expected_result_type=bytes)
        self.assertEqual(Integer(42).to_byte_array(), result)

    def test_storage_put_with_context(self):
        path = self.get_contract_path('StoragePutWithContext.py')

        engine = TestEngine()
        stored_value = b'\x01\x02\x03'
        result = self.run_smart_contract(engine, path, 'Main', b'test1')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2')
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'Main', b'test2', fake_storage={})
        self.assertIsVoid(result)
        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(b'test2', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

    def test_storage_delete_with_context(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.SYSCALL    # context = get_context()
            + Interop.StorageGetContext.interop_method_hash
            + Opcode.STLOC0
            + Opcode.LDARG0     # delete(key, context)
            + Opcode.LDLOC0
            + Opcode.SYSCALL
            + Interop.StorageDelete.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('StorageDeleteWithContext.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'example')
        self.assertIsVoid(result)
        self.assertIsNone(engine.storage_get(b'example', path))

        storage = {('example', path): 23}
        result = self.run_smart_contract(engine, path, 'Main', 'example', fake_storage=storage)
        self.assertIsVoid(result)
        self.assertIsNone(engine.storage_get('example', path))
        self.assertIsNone(engine.storage_get(b'example', path))

    def test_storage_find_with_context(self):
        path = self.get_contract_path('StorageFindWithContext.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'example')
        self.assertEqual([], result)

        storage = {('example_0', path): '0',
                   ('example_1', path): '1',
                   ('example_2', path): '3'}
        expected_result = [[key, value] for (key, sc), value in storage.items()]

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'example', fake_storage=storage)
        self.assertEqual(expected_result, result)

    def test_storage_find_with_options(self):
        path = self.get_contract_path('StorageFindWithOptions.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'example')
        self.assertEqual([], result)

        prefix = 'example'
        expected_result = [['_0', '0'],
                           ['_1', '1'],
                           ['_2', '2']
                           ]
        storage = {(prefix + key, path): value for (key, value) in expected_result}

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix, fake_storage=storage)
        self.assertEqual(expected_result, result)

    def test_boa2_storage_test(self):
        path = self.get_contract_path('StorageBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'sget', 'something', 'blah')
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'', result)

        result = self.run_smart_contract(engine, path, 'main', 'sput', 'something', 'blah')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 'sget', 'something', 'blah')
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'blah', result)

        result = self.run_smart_contract(engine, path, 'main', 'sdel', 'something', 'blah')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 'sget', 'something', 'blah')
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'', result)

    def test_boa2_storage_test2(self):
        path = self.get_contract_path('StorageBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'sget', 100, 10000000000)
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'', result)

        result = self.run_smart_contract(engine, path, 'main', 'sput', 100, 10000000000)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 'sget', 100, 10000000000)
        if isinstance(result, bytes):
            result = Integer.from_bytes(result)
        self.assertEqual(10000000000, result)

        result = self.run_smart_contract(engine, path, 'main', 'sdel', 100, 10000000000)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 'sget', 100, 10000000000)
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'', result)

    def test_storage_between_contracts(self):
        path1 = self.get_contract_path('StorageGetAndPut1.py')
        path2 = self.get_contract_path('StorageGetAndPut2.py')

        key = 'example_key'
        value = 42
        engine = TestEngine()

        result = self.run_smart_contract(engine, path1, 'put_value', key, value)
        self.assertIsVoid(result)

        storage_value = engine.storage_get(key, path1)
        self.assertIsNotNone(storage_value)
        self.assertEqual(value, Integer.from_bytes(storage_value))

        result = self.run_smart_contract(engine, path2, 'get_value', key)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path1, 'get_value', key)
        self.assertEqual(value, result)

    def test_create_map(self):
        path = self.get_contract_path('StorageCreateMap.py')

        map_key = b'example_'
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_from_map', 'test1',
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)

        stored_value = b'123'
        result = self.run_smart_contract(engine, path, 'insert_to_map', 'test1', stored_value)
        self.assertIsVoid(result)

        storage_value = engine.storage_get(b'test1', path)
        self.assertIsNone(storage_value)
        storage_value = engine.storage_get(map_key + b'test1', path)
        self.assertIsNotNone(storage_value)
        self.assertEqual(stored_value, storage_value)

        result = self.run_smart_contract(engine, path, 'get_from_map', 'test1',
                                         expected_result_type=bytes)
        self.assertEqual(stored_value, result)

        result = self.run_smart_contract(engine, path, 'delete_from_map', 'test1')
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_from_map', 'test1',
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)

    def test_import_storage(self):
        path = self.get_contract_path('ImportStorage.py')
        engine = TestEngine()

        prefix = 'unit'
        key = f'{prefix}_test'
        value = 1234

        result = self.run_smart_contract(engine, path, 'get_value', key)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix)
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'put_value', key, value)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_value', key)
        self.assertEqual(value, result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix)
        self.assertEqual([[key, Integer(value).to_byte_array()]], result)

        result = self.run_smart_contract(engine, path, 'delete_value', key)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_value', key)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'prefix')
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix)
        self.assertEqual([], result)

    def test_import_interop_storage(self):
        path = self.get_contract_path('ImportInteropStorage.py')
        engine = TestEngine()

        prefix = 'unit'
        key = f'{prefix}_test'
        value = 1234

        result = self.run_smart_contract(engine, path, 'get_value', key)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix)
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'put_value', key, value)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_value', key)
        self.assertEqual(value, result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix)
        self.assertEqual([[key, Integer(value).to_byte_array()]], result)

        result = self.run_smart_contract(engine, path, 'delete_value', key)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_value', key)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', 'prefix')
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'find_by_prefix', prefix)
        self.assertEqual([], result)

    def test_as_read_only(self):
        path = self.get_contract_path('StorageAsReadOnly.py')
        engine = TestEngine()

        key = 'key'
        value_old = 'old value'
        value_new = 'new value'

        # Putting old value in the storage
        result = self.run_smart_contract(engine, path, 'put_value_in_storage', key, value_old)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_value_in_storage_read_only', key)
        self.assertEqual(value_old, result)

        result = self.run_smart_contract(engine, path, 'get_value_in_storage', key)
        self.assertEqual(value_old, result)

        # Trying to put a new value in the storage using read_only won't work
        with self.assertRaisesRegex(TestExecutionException, self.VALUE_DOES_NOT_FALL_WITHIN_EXPECTED_RANGE_MSG):
            self.run_smart_contract(engine, path, 'put_value_in_storage_read_only', key, value_new)

        result = self.run_smart_contract(engine, path, 'get_value_in_storage_read_only', key)
        self.assertEqual(value_old, result)

        result = self.run_smart_contract(engine, path, 'get_value_in_storage', key)
        self.assertEqual(value_old, result)



from typing import Dict

from boa3.boa3 import Boa3
from boa3.compiler.compiler import Compiler
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestVariable(BoaTest):
    default_folder: str = 'test_sc/variable_test'

    def test_declaration_with_type(self):
        path = self.get_contract_path('DeclarationWithType.py')

        test_variable_id = 'a'
        test_method_id = 'Main'
        compiler = Compiler()

        expected_compiler_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.RET        # return
        )
        compiler_output = compiler.compile(path)
        self.assertEqual(expected_compiler_output, compiler_output)

        main_symbol_table: Dict[str, ISymbol] = self.get_compiler_analyser(compiler).symbol_table
        # the variable is local to a method, so it shouldn't be in the main symbol table
        self.assertFalse(test_variable_id in main_symbol_table)

        self.assertTrue(test_method_id in main_symbol_table)
        self.assertIsInstance(main_symbol_table[test_method_id], Method)
        method: Method = main_symbol_table[test_method_id]

        method_symbol_table: Dict[str, Variable] = method.symbols
        # the variable is local to this method, so it should be in the method symbol table
        self.assertTrue(test_variable_id in method_symbol_table)

    def test_declaration_without_type(self):
        path = self.get_contract_path('DeclarationWithoutType.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_assignment_with_type(self):
        input = 'unit_test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # assignment value
            + Integer(len(byte_input)).to_byte_array(min_length=1)
            + byte_input
            + Opcode.STLOC0     # variable address
            + Opcode.RET
        )

        path = self.get_contract_path('AssignmentWithType.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_assignment_without_type(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1      # assignment value
            + Opcode.STLOC0     # variable address
            + Opcode.RET
        )

        path = self.get_contract_path('AssignmentWithoutType.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_argument_assignment(self):
        input = 'unit_test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.INITSLOT         # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.PUSHDATA1      # assignment value
            + Integer(len(byte_input)).to_byte_array(min_length=1)
            + byte_input
            + Opcode.STARG0         # variable address
            + Opcode.RET
        )

        path = self.get_contract_path('ArgumentAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_multiple_assignments(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSHT      # a = b = c = True
            + Opcode.DUP            # c = True
            + Opcode.STLOC2
            + Opcode.DUP            # b = True
            + Opcode.STLOC1
            + Opcode.STLOC0         # a = True
            + Opcode.RET        # return
        )

        path = self.get_contract_path('MultipleAssignments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_multiple_assignments_set_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH2      # c = a[2] = b = 2
            + Opcode.DUP            # b = 2
            + Opcode.STLOC2
            + Opcode.LDLOC0         # a[2] = 2
            + Opcode.PUSH2
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SETITEM
            + Opcode.STLOC1         # c = 2
            + Opcode.RET        # return
        )

        path = self.get_contract_path('MultipleAssignmentsSetSequence.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_multiple_assignments_set_sequence_last(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH2      # a[2] = c = b = 2
            + Opcode.DUP            # b = 2
            + Opcode.STLOC2
            + Opcode.DUP            # c = 2
            + Opcode.STLOC1
            + Opcode.PUSH2          # a[2] = 2
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.LDLOC0
            + Opcode.REVERSE3
            + Opcode.SETITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('MultipleAssignmentsSetSequenceLast.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_multiple_assignments_mismatched_type(self):
        string = String('str').to_bytes()
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSHDATA1      # c = 'str'
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.STLOC0
            + Opcode.PUSHT      # a = b = c = True
            + Opcode.DUP            # c = True
            + Opcode.STLOC0
            + Opcode.DUP            # b = True
            + Opcode.STLOC2
            + Opcode.STLOC1         # a = True
            + Opcode.RET        # return
        )

        path = self.get_contract_path('MismatchedTypeMultipleAssignments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_multiple_assignments(self):
        path = self.get_contract_path('AssignmentWithTuples.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_many_assignments(self):
        expected_output = (
            Opcode.INITSLOT         # function signature
            + b'\x08'
            + b'\x00'
            + Opcode.PUSH0          # function body
            + Opcode.STLOC0
            + Opcode.PUSH1
            + Opcode.STLOC1
            + Opcode.PUSH2
            + Opcode.STLOC2
            + Opcode.PUSH3
            + Opcode.STLOC3
            + Opcode.PUSH4
            + Opcode.STLOC4
            + Opcode.PUSH5
            + Opcode.STLOC5
            + Opcode.PUSH6
            + Opcode.STLOC6
            + Opcode.PUSH7
            + Opcode.STLOC          # variable index greater than 6 uses another opcode
            + b'\x07'
            + Opcode.RET
        )

        path = self.get_contract_path('ManyAssignments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_return_arg_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # variable address
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnArgument.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', -140)
        self.assertEqual(-140, result)

    def test_return_local_var_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH1
            + Opcode.STLOC0
            + Opcode.PUSH1      # variable address
            + Opcode.RET
        )

        path = self.get_contract_path('ReturnLocalVariable.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', -140)
        self.assertEqual(1, result)

    def test_assign_local_with_arg_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0     # variable address
            + Opcode.RET
        )

        path = self.get_contract_path('AssignLocalWithArgument.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', -140)
        self.assertEqual(-140, result)

    def test_using_undeclared_variable(self):
        path = self.get_contract_path('UsingUndeclaredVariable.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_return_undeclared_variable(self):
        path = self.get_contract_path('ReturnUndeclaredVariable.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_assign_value_mismatched_type(self):
        string_value = '1'
        byte_input = String(string_value).to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = '1'
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeAssignValue.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_assign_binary_operation_mismatched_type(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3  # a = 1 + 2
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeAssignBinOp.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_assign_unary_operation_mismatched_type(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5  # a = -5
            + Opcode.NEGATE
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeAssignUnOp.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_assign_mixed_operations_mismatched_type(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x03'
            + Opcode.LDARG1  # b = min <= a - 2 <= max
            + Opcode.LDARG0
            + Opcode.PUSH2
            + Opcode.SUB
            + Opcode.LE
            + Opcode.LDARG0
            + Opcode.PUSH2
            + Opcode.SUB
            + Opcode.LDARG2
            + Opcode.LE
            + Opcode.BOOLAND
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeAssignMixedOp.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_assign_sequence_get_mismatched_type(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0  # b = a[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeAssignSequenceGet.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_assign_sequence_set_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeAssignSequenceSet.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_aug_assign_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeAugAssign.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_invalid_type_format_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeInvalidTypeFormat.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_global_declaration_with_assignment(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH10
            + Opcode.STSFLD0    # a = 10
            + Opcode.RET
        )
        path = self.get_contract_path('GlobalDeclarationWithArgumentWrittenAfter.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(10, result)

    def test_global_declaration_without_assignment(self):
        path = self.get_contract_path('GlobalDeclarationWithoutAssignment.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_global_assignment_with_type(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH10
            + Opcode.STSFLD0    # a = 10
            + Opcode.RET
        )
        path = self.get_contract_path('GlobalAssignmentWithType.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(10, result)

    def test_global_assignment_without_type(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH10
            + Opcode.STSFLD0    # a = 10
            + Opcode.RET
        )
        path = self.get_contract_path('GlobalAssignmentWithoutType.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(10, result)

    def test_global_tuple_multiple_assignments(self):
        path = self.get_contract_path('GlobalAssignmentWithTuples.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_global_chained_multiple_assignments(self):
        path = self.get_contract_path('GlobalMultipleAssignments.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_a')
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'get_c')
        self.assertEqual(15, result)

        result = self.run_smart_contract(engine, path, 'set_a', 100)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'get_a')
        self.assertEqual(100, result)

    def test_many_global_assignments(self):
        expected_output = (
            Opcode.LDSFLD + b'\x07'
            + Opcode.LDSFLD6    # [a, b, c, d, e, f, g, h]
            + Opcode.LDSFLD5
            + Opcode.LDSFLD4
            + Opcode.LDSFLD3
            + Opcode.LDSFLD2
            + Opcode.LDSFLD1
            + Opcode.LDSFLD0
            + Opcode.PUSH8
            + Opcode.PACK       # return [a, b, c, d, e, f, g, h]
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x08'           # number of globals
            + Opcode.PUSH0      # a = 0
            + Opcode.STSFLD0
            + Opcode.PUSH1      # b = 1
            + Opcode.STSFLD1
            + Opcode.PUSH2      # c = 2
            + Opcode.STSFLD2
            + Opcode.PUSH3      # d = 3
            + Opcode.STSFLD3
            + Opcode.PUSH4      # e = 4
            + Opcode.STSFLD4
            + Opcode.PUSH5      # f = 5
            + Opcode.STSFLD5
            + Opcode.PUSH6      # g = 6
            + Opcode.STSFLD6
            + Opcode.PUSH7      # h = 7
            + Opcode.STSFLD + b'\x07'   # variable index greater than 6 uses another opcode
            + Opcode.RET
        )

        path = self.get_contract_path('ManyGlobalAssignments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1, 2, 3, 4, 5, 6, 7], result)

    def test_list_global_assignment(self):
        path = self.get_contract_path('ListGlobalAssignment.py')
        engine = TestEngine()

        expected_value = [1, 2, 3, 4]
        result = self.run_smart_contract(engine, path, 'get_from_global')
        self.assertEqual(expected_value, result)

        result = self.run_smart_contract(engine, path, 'get_from_class')
        self.assertEqual(expected_value, result)

        result = self.run_smart_contract(engine, path, 'get_from_class_without_assigning')
        self.assertEqual([], result)

    def test_global_assignment_between_functions(self):
        expected_output = (
            Opcode.LDSFLD0
            + Opcode.RET
            + Opcode.LDSFLD1
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x02'           # number of globals
            + Opcode.PUSH10     # a = 10
            + Opcode.STSFLD0
            + Opcode.PUSH5      # b = 5
            + Opcode.STSFLD1
            + Opcode.RET
        )
        path = self.get_contract_path('GlobalAssignmentBetweenFunctions.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'example')
        self.assertEqual(5, result)

    def test_global_variable_in_class_method(self):
        path = self.get_contract_path('GlobalVariableInClassMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'use_variable_in_func')
        self.assertEqual(42, result)

        result = self.run_smart_contract(engine, path, 'use_variable_in_map')
        self.assertEqual({'val1': 1, 'val2': 2, 'bar': 42}, result)

    def test_get_global_variable_value_written_after(self):
        expected_output = (
            Opcode.LDSFLD + b'\x07'
            + Opcode.LDSFLD6    # [a, b, c, d, e, f, g, h]
            + Opcode.LDSFLD5
            + Opcode.LDSFLD4
            + Opcode.LDSFLD3
            + Opcode.LDSFLD2
            + Opcode.LDSFLD1
            + Opcode.LDSFLD0
            + Opcode.PUSH8
            + Opcode.PACK       # return [a, b, c, d, e, f, g, h]
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x08'           # number of globals
            + Opcode.PUSH0      # a = 0
            + Opcode.STSFLD0
            + Opcode.PUSH1      # b = 1
            + Opcode.STSFLD1
            + Opcode.PUSH2      # c = 2
            + Opcode.STSFLD2
            + Opcode.PUSH3      # d = 3
            + Opcode.STSFLD3
            + Opcode.PUSH4      # e = 4
            + Opcode.STSFLD4
            + Opcode.PUSH5      # f = 5
            + Opcode.STSFLD5
            + Opcode.PUSH6      # g = 6
            + Opcode.STSFLD6
            + Opcode.PUSH7      # h = 7
            + Opcode.STSFLD + b'\x07'   # variable index greater than 6 uses another opcode
            + Opcode.RET
        )
        path = self.get_contract_path('GetGlobalValueWrittenAfter.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1, 2, 3, 4, 5, 6, 7], result)

    def test_assign_local_shadowing_global_with_arg_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # b = a  // this b is not the global b
            + Opcode.STLOC0
            + Opcode.LDLOC0     # variable address
            + Opcode.RET
            + Opcode.INITSSLOT  # global variables
            + b'\x01'           # number of globals
            + Opcode.PUSH0      # b = 0
            + Opcode.STSFLD0
            + Opcode.RET
        )
        path = self.get_contract_path('AssignLocalWithArgumentShadowingGlobal.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', -140)
        self.assertEqual(-140, result)

    def test_assign_global_in_function_with_global_keyword(self):
        path = self.get_contract_path('GlobalAssignmentInFunctionWithArgument.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_b')
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'get_b')
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'Main', -140)
        self.assertEqual(-140, result)

        result = self.run_smart_contract(engine, path, 'get_b')
        self.assertEqual(-140, result)

    def test_assign_void_function_call(self):
        path = self.get_contract_path('AssignVoidFunctionCall.py')
        engine = TestEngine()

        output = Boa3.compile(path)
        self.assertIn(Opcode.NOP, output)

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(None, result)

    def test_anonymous_variable(self):
        path = self.get_contract_path('AnonymousVariable')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(400, result)

    def test_assign_starred_variable(self):
        path = self.get_contract_path('AssignStarredVariable.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_variables_in_different_scope_with_same_name(self):
        path = self.get_contract_path('DifferentScopesWithSameName.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test')
        self.assertEqual(1_000, result)

    def test_instance_variable_and_variable_with_same_name(self):
        path = self.get_contract_path('InstanceVariableAndVariableWithSameName.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'test')
        self.assertEqual([10], result)

    def test_inner_object_variable_access(self):
        path = self.get_contract_path('InnerObjectVariableAccess.py')
        engine = TestEngine()

        expected_return = 'InnerObjectVariableAccess'
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_return, result)



from boa3.builtin.interop.role.roletype import Role
from boa3.builtin.type import ECPoint


def get_designated_by_role(role: Role, index: int) -> ECPoint:
    """
    Gets the list of nodes for the specified role.

    :param role: the type of the role
    :type role: Role
    :param index: the index of the block to be queried
    :type index: int

    :return: the public keys of the nodes
    :rtype: ECPoint
    """
    pass



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple

from boa3 import constants
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.type.classes.classarraytype import ClassArrayType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class TransactionType(ClassArrayType):
    """
    A class used to represent Neo Transaction class
    """

    def __init__(self):
        super().__init__('Transaction')
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.model.type.collection.sequence.uint256type import UInt256Type

        self._variables: Dict[str, Variable] = {
            'hash': Variable(UInt256Type.build()),
            'version': Variable(Type.int),
            'nonce': Variable(Type.int),
            'sender': Variable(UInt160Type.build()),
            'system_fee': Variable(Type.int),
            'network_fee': Variable(Type.int),
            'valid_until_block': Variable(Type.int),
            'script': Variable(Type.bytes),
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        # was having a problem with recursive import
        if self._constructor is None:
            self._constructor: Method = TransactionMethod(self)
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> TransactionType:
        if value is None or cls._is_type_of(value):
            return _Transaction

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, TransactionType)


_Transaction = TransactionType()


class TransactionMethod(IBuiltinMethod):

    def __init__(self, return_type: TransactionType):
        identifier = '-Transaction__init__'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 0

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer

        uint160_default = Integer(constants.SIZE_OF_INT160).to_byte_array() + bytes(constants.SIZE_OF_INT160)
        uint256_default = Integer(constants.SIZE_OF_INT256).to_byte_array() + bytes(constants.SIZE_OF_INT256)

        return [
            (Opcode.PUSHDATA1, Integer(0).to_byte_array()),  # script
            (Opcode.PUSH0, b''),  # valid_until_block
            (Opcode.PUSH0, b''),  # network_fee
            (Opcode.PUSH0, b''),  # system_fee
            (Opcode.PUSHDATA1, uint160_default),  # sender
            (Opcode.PUSH0, b''),  # nonce
            (Opcode.PUSH0, b''),  # version
            (Opcode.PUSHDATA1, uint256_default),  # hash
            (Opcode.PUSH8, b''),
            (Opcode.PACK, b'')
        ]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



from boa3.boa3 import Boa3
from boa3.exception import CompilerError, CompilerWarning
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestDict(BoaTest):
    default_folder: str = 'test_sc/dict_test'

    def test_dict_int_keys(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {1: 15, 2: 14, 3: 13}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = 15
            + Opcode.PUSH15
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 14
            + Opcode.PUSH14
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 13
            + Opcode.PUSH13
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('IntKeyDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_str_keys(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('StrKeyDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_any_value(self):
        nine = String('nine').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {1: True, 2: 4, 3: 'nine'}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = True
            + Opcode.PUSHT
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 4
            + Opcode.PUSH4
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 'nine'
            + Opcode.PUSHDATA1
            + Integer(len(nine)).to_byte_array(min_length=1)
            + nine
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('AnyValueDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_of_dicts(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = map
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = {14: False, 12: True, 5: True}
            + Opcode.NEWMAP
            + Opcode.DUP
            + Opcode.PUSH14
            + Opcode.PUSHF
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH12
            + Opcode.PUSHT
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH5
            + Opcode.PUSHT
            + Opcode.SETITEM
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = {0: True, 6: False}
            + Opcode.NEWMAP
            + Opcode.DUP
            + Opcode.PUSH0
            + Opcode.PUSHT
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH6
            + Opcode.PUSHF
            + Opcode.SETITEM
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = {11: False}
            + Opcode.NEWMAP
            + Opcode.DUP
            + Opcode.PUSH11
            + Opcode.PUSHF
            + Opcode.SETITEM
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('DictOfDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_assign_empty_dict(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {}
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('EmptyDictAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_type_hint_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.NEWMAP  # a = {1: 15, 2: 14, 3: 13}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = 15
            + Opcode.PUSH15
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 14
            + Opcode.PUSH14
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 13
            + Opcode.PUSH13
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('TypeHintAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_variable_keys_and_values(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH1   # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2   # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH3   # c = 3
            + Opcode.STLOC2
            + Opcode.NEWMAP  # d = {a: c, b: a, c: b}
            + Opcode.DUP
            + Opcode.PUSH1      # map[a] = c
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[b] = a
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[c] = b
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.STLOC3
            + Opcode.RET
        )

        path = self.get_contract_path('VariableDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_dict_get_value(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH0
            + Opcode.PICKITEM
            + Opcode.RET
        )

        path = self.get_contract_path('GetValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', {0: 'zero'})
        self.assertEqual('zero', result)

        with self.assertRaisesRegex(TestExecutionException, self.MAP_KEY_NOT_FOUND_ERROR_MSG):
            self.run_smart_contract(engine, path, 'Main', {1: 'one'})

    def test_dict_get_value_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeGetValue.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_dict_set_value(self):
        ok = String('ok').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH0
            + Opcode.PUSHDATA1
            + Integer(len(ok)).to_byte_array(min_length=1)
            + ok
            + Opcode.SETITEM
            + Opcode.LDARG0
            + Opcode.RET
        )

        path = self.get_contract_path('SetValue.py')
        output = self.assertCompilerNotLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', {0: 'zero'})
        self.assertEqual({0: 'ok'}, result)
        result = self.run_smart_contract(engine, path, 'Main', {1: 'one'})
        self.assertEqual({0: 'ok', 1: 'one'}, result)

    def test_dict_set_value_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeSetValue.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_dict_keys(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.LDLOC0  # b = a.keys()
            + Opcode.KEYS
            + Opcode.STLOC1
            + Opcode.LDLOC1  # return b
            + Opcode.RET
        )

        path = self.get_contract_path('KeysDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(['one', 'two', 'three'], result)

    def test_dict_keys_mismatched_type(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.LDLOC0  # b = a.keys()
            + Opcode.KEYS
            + Opcode.STLOC1
            + Opcode.LDLOC1  # return b
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeKeysDict.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(['one', 'two', 'three'], result)

    def test_dict_values(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.LDLOC0  # b = a.values()
            + Opcode.VALUES
            + Opcode.STLOC1
            + Opcode.LDLOC1  # return b
            + Opcode.RET
        )

        path = self.get_contract_path('ValuesDict.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([1, 2, 3], result)

    def test_dict_values_mismatched_type(self):
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.NEWMAP  # a = {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
            + Opcode.STLOC0
            + Opcode.LDLOC0  # b = a.values()
            + Opcode.VALUES
            + Opcode.STLOC1
            + Opcode.LDLOC1  # return b
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedTypeValuesDict.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([1, 2, 3], result)

    def test_dict_boa2_test2(self):
        path = self.get_contract_path('DictBoa2Test2.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(7, result)

    def test_dict_any_key_and_value(self):
        path = self.get_contract_path('DictAnyKeyAndValue.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(66, result)

    def test_boa2_dict_test1(self):
        path = self.get_contract_path('DictBoa2Test1.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        from typing import Dict
        self.assertIsInstance(result, Dict)

    def test_boa2_dict_test3(self):
        path = self.get_contract_path('DictBoa2Test3.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        from typing import Dict
        self.assertIsInstance(result, Dict)
        self.assertEqual(result, {})

    def test_boa2_dict_test4(self):
        path = self.get_contract_path('DictBoa2Test4.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(10, result)

    def test_boa2_dict_test5_should_not_compile(self):
        # this doesn't compile in boa2, but should compile here
        path = self.get_contract_path('DictBoa2Test5ShouldNotCompile.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(result, {'a': 2})

    def test_boa2_dict_test6_should_not_compile(self):
        # this doesn't compile in boa2, but should compile here
        path = self.get_contract_path('DictBoa2Test6ShouldNotCompile.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(result, {'a': 1, 'b': 2})

    def test_boa2_dict_test_keys(self):
        path = self.get_contract_path('DictBoa2TestKeys.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('abblahmzmcallltrs', result)

    def test_boa2_dict_test_values(self):
        path = self.get_contract_path('DictBoa2TestValues.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(55, result)

    def test_dict_pop(self):
        path = self.get_contract_path('DictPop.py')
        engine = TestEngine()

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'a'
        result = self.run_smart_contract(engine, path, 'main', dict_, key)
        value = dict_.pop(key)
        self.assertEqual((dict_, value), tuple(result))

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'd'
        result = self.run_smart_contract(engine, path, 'main', dict_, key)
        value = dict_.pop(key)
        self.assertEqual((dict_, value), tuple(result))

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'key not inside'
        with self.assertRaisesRegex(TestExecutionException, self.MAP_KEY_NOT_FOUND_ERROR_MSG):
            self.run_smart_contract(engine, path, 'main', dict_, key)

    def test_dict_pop_default(self):
        path = self.get_contract_path('DictPopDefault.py')
        engine = TestEngine()

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'a'
        default = 'test'
        result = self.run_smart_contract(engine, path, 'main', dict_, key, default)
        value = dict_.pop(key, default)
        self.assertEqual((dict_, value), tuple(result))

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'd'
        default = 'test'
        result = self.run_smart_contract(engine, path, 'main', dict_, key, default)
        value = dict_.pop(key, default)
        self.assertEqual((dict_, value), tuple(result))

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'not inside'
        default = 'test'
        result = self.run_smart_contract(engine, path, 'main', dict_, key, default)
        value = dict_.pop(key, default)
        self.assertEqual((dict_, value), tuple(result))

        dict_ = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        key = 'not inside'
        default = 123456
        result = self.run_smart_contract(engine, path, 'main', dict_, key, default)
        value = dict_.pop(key, default)
        self.assertEqual((dict_, value), tuple(result))

    def test_dict_copy(self):
        path = self.get_contract_path('DictCopy.py')
        engine = TestEngine()

        _dict = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5}
        result = self.run_smart_contract(engine, path, 'copy_dict', _dict)
        self.assertEqual(2, len(result))
        self.assertEqual({'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5}, result[0])
        self.assertEqual({'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'unit': 'test'}, result[1])

    def test_dict_copy_builtin_call(self):
        path = self.get_contract_path('CopyDictBuiltinCall.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'copy_dict', {0: 10, 1: 11, 2: 12}, 3, 13)
        self.assertEqual({0: 10, 1: 11, 2: 12}, result[0])
        self.assertEqual({0: 10, 1: 11, 2: 12, 3: 13}, result[1])

        result = self.run_smart_contract(engine, path, 'copy_dict', {'dict': 1, 'unit': 2, 'test': 3}, 'copy', 4)
        self.assertEqual({'dict': 1, 'unit': 2, 'test': 3}, result[0])
        self.assertEqual({'dict': 1, 'unit': 2, 'test': 3, 'copy': 4}, result[1])

        result = self.run_smart_contract(engine, path, 'copy_dict', {True: 1, False: 0}, True, 99)
        self.assertEqual({True: 1, False: 0}, result[0])
        self.assertEqual({True: 99, False: 0}, result[1])



from typing import Any, Dict

from boa3.builtin.compile_time import public
from package_with_import import Module as imported_module


@public
def get_token_json(token_id: bytes) -> Dict[str, Any]:
    return imported_module.get_token_json(token_id)  # this method doesn't exist in the imported module



from typing import List, Dict, Optional

from boa3.neo.vm.VMCode import VMCode


class VMCodeMap:
    def __init__(self):
        self._vm_code_list: List[VMCode] = []
        self._vm_code_addresses: List[int] = []

        # optimization so it's not needed to iterate over everything in search of targets
        self._vm_code_with_target: List[VMCode] = []

    def __len__(self) -> int:
        return self._vm_code_list.__len__()

    def clear(self):
        self._vm_code_addresses.clear()
        self._vm_code_list.clear()

    def get_code_map(self) -> Dict[int, VMCode]:
        size = len(self)
        return {self._vm_code_addresses[index]: self._vm_code_list[index] for index in range(size)}

    def get_code_list(self) -> List[VMCode]:
        return self._vm_code_list

    def get_code_with_target_list(self) -> List[VMCode]:
        return self._vm_code_with_target

    def get_bytecode_size(self) -> int:
        if len(self) < 1:
            return 0

        return self._vm_code_addresses[-1] + self._vm_code_list[-1].size

    def insert_code(self, vm_code: VMCode, has_target: bool = False):
        if vm_code not in self._vm_code_list:
            self._vm_code_addresses.append(self.get_bytecode_size())
            self._vm_code_list.append(vm_code)

            if has_target:
                self._vm_code_with_target.append(vm_code)

    def get_code(self, address: int) -> Optional[VMCode]:
        try:
            index = self._vm_code_addresses.index(address)
        except ValueError:
            # the address is not int the list
            if address >= self.get_bytecode_size():
                # the address is not in the bytecode
                return None

            # if the address is not the start of a instruction, gets the last instruction before given address
            code_address = 0
            for addr in self._vm_code_addresses:
                if addr > address:
                    break
                code_address = addr

            index = self._vm_code_addresses.index(code_address)

        return self._vm_code_list[index]

    def get_start_address(self, vm_code: VMCode) -> int:
        try:
            index = self._vm_code_list.index(vm_code)
            return self._vm_code_addresses[index]
        except ValueError:
            return 0

    def get_end_address(self, vm_code: VMCode) -> int:
        try:
            index = self._vm_code_list.index(vm_code) + 1
            if index == len(self._vm_code_list):
                return self.get_bytecode_size()
            else:
                return self._vm_code_addresses[index] - 1
        except ValueError:
            return 0

    def get_addresses(self, start_address: int, end_address: int) -> List[int]:
        if start_address > end_address:
            start_address, end_address = end_address, start_address

        addresses = []
        for address in range(start_address, end_address + 1):
            if address in self._vm_code_addresses:
                addresses.append(address)
        return addresses

    def get_addresses_from_codes(self, codes: List[VMCode]) -> List[int]:
        if len(codes) < 1:
            return []

        addresses = []
        for vm_code in codes:
            try:
                index = self._vm_code_list.index(vm_code)
                addresses.append(self._vm_code_addresses[index])
            except ValueError:
                continue

        return addresses

    def get_opcodes(self, addresses: List[int]) -> List[VMCode]:
        codes = []

        for address in sorted(addresses):
            try:
                index = self._vm_code_addresses.index(address)
                codes.append(self._vm_code_list[index])
            except ValueError:
                # address not in list
                continue

        return codes

    def update_addresses(self, start_address: int = 0):
        next_address = -1
        final_size = len(self._vm_code_list)

        if len(self._vm_code_addresses) > final_size:
            self._vm_code_addresses = self._vm_code_addresses[:final_size]

        for index in range(final_size):
            address = self._vm_code_addresses[index]

            if address >= start_address:
                if next_address < 0:
                    if index > 0:
                        new_address = self._vm_code_addresses[index - 1]
                        next_address = new_address + self._vm_code_list[index - 1].size
                    else:
                        next_address = 0

                if next_address != address:
                    if index < len(self._vm_code_addresses):
                        self._vm_code_addresses[index] = next_address
                    else:
                        self._vm_code_addresses.append(next_address)

                next_address += self._vm_code_list[index].size

    def move_to_end(self, first_code_address: int, last_code_address: int) -> Optional[int]:
        if last_code_address < first_code_address:
            return

        if (len(self._vm_code_addresses) > 0 and
                last_code_address == self._vm_code_addresses[-1]):
            # there's nothing to change if it's moving the all the codes
            return

        first_index = -1
        last_index = 0
        for index, address in enumerate(self._vm_code_addresses):
            if first_code_address <= address and first_index < 0:
                first_index = index
            elif address > last_code_address:
                last_index = index
                break

        if first_index >= 0:
            # if the first index was not set, there's nothing to move
            if last_index < first_index:
                last_index = len(self._vm_code_addresses)

            self._vm_code_list[first_index:] = (self._vm_code_list[last_index:] +
                                                self._vm_code_list[first_index:last_index])
            self.update_addresses(first_code_address)

        index = self.get_bytecode_size()
        return index

    def remove_opcodes_by_addresses(self, addresses: List[int]):
        was_changed = False
        # reversed so we only need to update addresses once after all are removed
        for code_address in sorted(addresses, reverse=True):
            try:
                index = self._vm_code_addresses.index(code_address)
                code = self._vm_code_list.pop(index)

                was_changed = True
                self._vm_code_with_target.remove(code)
            except ValueError:
                # don't stop the loop if an address is not found
                continue

        if was_changed:
            self.update_addresses(min(addresses))



from __future__ import annotations

from typing import Dict, List, Optional, Union

from boa3.compiler.codegenerator.methodtokencollection import MethodTokenCollection
from boa3.compiler.codegenerator.vmcodemap import VMCodeMap
from boa3.compiler.compileroutput import CompilerOutput
from boa3.model.builtin.method import IBuiltinMethod
from boa3.neo.vm.VMCode import VMCode
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.OpcodeInformation import OpcodeInformation
from boa3.neo3.contracts.contracttypes import CallFlags


class VMCodeMapping:
    """
    This class is responsible for managing the Neo VM instruction during the bytecode generation.
    """
    _instance: VMCodeMapping = None

    @classmethod
    def instance(cls):
        """
        :return: the singleton instance
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self._code_map: VMCodeMap = VMCodeMap()
        self._method_tokens: MethodTokenCollection = MethodTokenCollection()

    @classmethod
    def reset(cls):
        """
        Resets the map to the first state
        """
        if cls._instance is not None:
            cls._instance._code_map.clear()
            cls._instance._method_tokens.clear()

    def add_method_token(self, method: IBuiltinMethod, call_flag: CallFlags) -> Optional[int]:
        """
        Creates a new method token if the method call another contract and return its id.
        Otherwise, returns None
        """
        if hasattr(method, 'contract_script_hash'):
            return self._method_tokens.append(method, call_flag)
        return None

    @property
    def codes(self) -> List[VMCode]:
        """
        Gets a list with the included vm codes

        :return: a list of vm codes ordered by its address in the bytecode
        """
        return self._code_map.get_code_list()

    @property
    def code_map(self) -> Dict[int, VMCode]:
        """
        Gets a dictionary that maps each vm code with its address.

        :return: a dictionary that maps each instruction with its address. The keys are ordered by the address.
        """
        return self._code_map.get_code_map()

    def targeted_address(self) -> Dict[int, List[int]]:
        """
        Gets a dictionary that maps each address to the opcodes that targets it

        :return: a dictionary that maps the targeted instructions to its source.
        """
        target_maps = {}
        for code in self._code_map.get_code_with_target_list():
            if code.target is not None and code.target is not code:
                address = self.get_start_address(code)
                target = self.get_start_address(code.target)
                if target not in target_maps:
                    target_maps[target] = [address]
                else:
                    target_maps[target].append(address)
        return target_maps

    def bytecode(self) -> bytes:
        """
        Gets the bytecode of the translated code

        :return: the generated bytecode
        """
        self._remove_empty_targets()
        self._update_larger_codes()

        bytecode = bytearray()
        for code in self.codes:
            bytecode += code.opcode
            if code.data is not None:
                bytecode += code.data
        return bytes(bytecode)

    def result(self) -> CompilerOutput:
        """
        Gets the complete output of the translated code
        """
        bytecode = self.bytecode()
        return CompilerOutput(bytecode, self._method_tokens.to_list())

    @property
    def bytecode_size(self) -> int:
        return self._code_map.get_bytecode_size()

    def insert_code(self, vm_code: VMCode):
        return self._code_map.insert_code(vm_code, has_target=OpcodeHelper.has_target(vm_code.opcode))

    def get_code(self, address: int) -> Optional[VMCode]:
        """
        Gets the VM Opcode at the given position

        :param address: the position of the opcode
        :return: the opcode if it exists. None otherwise
        :rtype: VMCode or None
        """
        return self._code_map.get_code(address)

    def get_addresses(self, start_address: int, end_address: int) -> List[int]:
        return self._code_map.get_addresses(start_address, end_address)

    def get_start_address(self, vm_code: VMCode) -> int:
        """
        Gets the vm code's first byte address

        :param vm_code: the instruction to get the address
        :return: the vm code's address if it's in the map. Otherwise, return's zero.
        """
        return self._code_map.get_start_address(vm_code)

    def get_end_address(self, vm_code: VMCode) -> int:
        """
        Gets the vm code's last byte address

        :param vm_code: the instruction to get the address
        :return: the vm code's last address if it's in the map. Otherwise, return's zero.
        """
        return self._code_map.get_end_address(vm_code)

    def get_opcodes(self, addresses: List[int]) -> List[VMCode]:
        return self._code_map.get_opcodes(addresses)

    def update_vm_code(self, vm_code: VMCode, opcode: OpcodeInformation, data: bytes = bytes()):
        """
        Updates the information from an inserted code

        :param vm_code: code to be updated
        :param opcode: updated opcode information
        :param data: updated opcode data
        """
        code_size = vm_code.size
        vm_code._info = opcode
        vm_code._data = data
        if vm_code.size != code_size:
            self._update_addresses(self.get_start_address(vm_code))

    def _update_addresses(self, start_address: int = 0):
        """
        Updates the instruction map's keys when a opcode is changed

        :param start_address: the address from the changed opcode
        """
        return self._code_map.update_addresses(start_address)

    def _update_targets(self):
        from boa3.neo.vm.type.Integer import Integer
        for code in self._code_map.get_code_with_target_list():
            if code.target is None:
                relative = Integer.from_bytes(code.data)
                absolute = self._code_map.get_start_address(code) + relative
                if absolute in self.code_map:
                    code.set_target(self.code_map[absolute])

    def _update_larger_codes(self):
        """
        Checks if each instruction data fits in its opcode maximum size and updates the opcode from those that don't
        """
        # gets a list with all instructions which its opcode has a larger equivalent, ordered by its address
        instr_with_small_codes = [code for code in self._code_map.get_code_list() if OpcodeHelper.has_larger_opcode(code.opcode)]
        instr_with_small_codes.sort(key=lambda code: self.get_start_address(code), reverse=True)

        from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
        # total_len is initialized with zero because the loop must run at least once
        total_len = 0
        current_size = self.bytecode_size

        # if any instruction is updated, the following instruction addresses and the total size will change as well
        # with the change, previous instruction data may have overflowed the opcode maximum value
        # to make sure, it must check the opcodes that haven't changed again
        while total_len != current_size:
            total_len = current_size

            # verifies each instruction data length
            for code in instr_with_small_codes.copy():  # it's a copy because the list may change during the iteration
                if len(code.raw_data) > code.info.max_data_len:
                    # gets the shortest opcode equivalent that fits the instruction data
                    info = OpcodeInfo.get_info(OpcodeHelper.get_larger_opcode(code.opcode))
                    while len(code.raw_data) > info.max_data_len and OpcodeHelper.has_larger_opcode(info.opcode):
                        info = OpcodeInfo.get_info(OpcodeHelper.get_larger_opcode(code.opcode))

                    self.update_vm_code(code, info)
                    if info.opcode == OpcodeHelper.get_larger_opcode(info.opcode):
                        # if it's the largest equivalent, it won't be updated anymore
                        instr_with_small_codes.remove(code)
            current_size = self.bytecode_size

    def _validate_targets(self, code_or_address: Union[int, VMCode]):
        if isinstance(code_or_address, int):
            address = code_or_address
            code = self.get_code(address)
        else:
            code = code_or_address
            address = self.get_start_address(code)

        targeted_addresses = self.targeted_address()

        if address in targeted_addresses:
            next_address = self.get_end_address(code) + 1
            if next_address < self.bytecode_size:
                next_code = self._code_map.get_code(next_address)
                for source in targeted_addresses[address]:
                    self._code_map.get_code(source).set_target(next_code)

    def move_to_end(self, first_code_address: int, last_code_address: int) -> int:
        """
        Moves a set of instructions to the end of the current bytecode

        :param first_code_address: first instruction start address
        :param last_code_address: last instruction end address
        """
        result = self._code_map.move_to_end(first_code_address, last_code_address)
        if not isinstance(result, int):
            return self.bytecode_size

        self._update_targets()
        return result

    def remove_opcodes(self, first_code_address: int, last_code_address: int = None):
        if not isinstance(last_code_address, int):
            last_code_address = self.bytecode_size
        addresses_to_remove = self._code_map.get_addresses(first_code_address, last_code_address)
        for address in addresses_to_remove:
            self._validate_targets(address)
        return self._code_map.remove_opcodes_by_addresses(addresses_to_remove)

    def remove_opcodes_by_code(self, codes: List[VMCode]):
        addresses_to_remove = self._code_map.get_addresses_from_codes(codes)
        for address in addresses_to_remove:
            self._validate_targets(address)
        return self._code_map.remove_opcodes_by_addresses(addresses_to_remove)

    def _remove_empty_targets(self):
        """
        Checks if each instruction that requires a target has one set and remove those that don't
        """
        addresses_to_remove = []
        for code in self._code_map.get_code_with_target_list():
            if code.target is None or code.target is code:
                address = self.get_start_address(code)
                self._validate_targets(address)
                addresses_to_remove.append(address)

        if len(addresses_to_remove) > 0:
            self._code_map.remove_opcodes_by_addresses(addresses_to_remove)



from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class InsertMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        if not isinstance(sequence_type, MutableSequenceType):
            from boa3.model.type.type import Type
            sequence_type = Type.mutableSequence

        self_arg = Variable(sequence_type)
        index_arg = Variable(sequence_type.valid_key)
        item_arg = Variable(sequence_type.value_type)

        identifier = 'insert'
        args: Dict[str, Variable] = {'self': self_arg,
                                     '__index': index_arg,
                                     '__object': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 3:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        sequence_type: IType = params[0].type
        index_type: IType = params[1].type
        value_type: IType = params[2].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        return sequence_type.key_type.is_type_of(index_type) and sequence_type.value_type.is_type_of(value_type)

    def validate_negative_arguments(self) -> List[int]:
        return [list(self.args).index('__index')]

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        return [
            # insert(pos, index)
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),     # array
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),     # array[-1]
            (Opcode.DEC, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.PUSH2, b''),    # array.append(array[-1])
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.APPEND, b''),   # for x in range(index + 1, len(array) - 1)
            (Opcode.SWAP, b''),     # x = len(array)
            (Opcode.JMP, Integer(16).to_byte_array(signed=True, min_length=1)),
            (Opcode.DEC, b''),      # x =- 1
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),     # aux = array[x]
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.OVER, b''),
            (Opcode.OVER, b''),
            (Opcode.INC, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.SETITEM, b''),      # array[x] = value
            (Opcode.DUP, b''),          # value = aux
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.GT, b''),
            (Opcode.JMPIF, Integer(-18).to_byte_array(signed=True, min_length=1)),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.SWAP, b''),     # array[index] = y
            (Opcode.SETITEM, b''),
        ]

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return InsertMethod(value)
        return super().build(value)



from typing import List, Union

from boa3.builtin.interop.blockchain.block import Block
from boa3.builtin.interop.blockchain.signer import Signer
from boa3.builtin.interop.blockchain.transaction import Transaction
from boa3.builtin.interop.blockchain.vmstate import VMState
from boa3.builtin.interop.contract import Contract
from boa3.builtin.type import UInt160, UInt256


def get_contract(hash: UInt160) -> Contract:
    """
    Gets a contract with a given hash.

    :param hash: a smart contract hash
    :type hash: UInt160
    :return: a contract
    :rtype: Contract

    :raise Exception: raised if hash length isn't 20 bytes.
    """
    pass


def get_block(index_or_hash: Union[int, UInt256]) -> Block:
    """
    Gets the block with the given index or hash.

    :param index_or_hash: index or hash identifier of the block
    :type index_or_hash: int or UInt256
    :return: the desired block, if exists. None otherwise
    :rtype: Block or None
    """
    pass


def get_transaction(hash_: UInt256) -> Transaction:
    """
    Gets a transaction with the given hash.

    :param hash_: hash identifier of the transaction
    :type hash_: UInt256
    :return: the Transaction, if exists. None otherwise
    """
    pass


def get_transaction_from_block(block_hash_or_height: Union[UInt256, int], tx_index: int) -> Transaction:
    """
    Gets a transaction from a block.

    :param block_hash_or_height: a block identifier
    :type block_hash_or_height: UInt256 or int
    :param tx_index: the transaction identifier in the block
    :type tx_index: int
    :return: the Transaction, if exists. None otherwise
    """
    pass


def get_transaction_height(hash_: UInt256) -> int:
    """
    Gets the height of a transaction.

    :param hash_: hash identifier of the transaction
    :type hash_: UInt256
    :return: height of the transaction
    """
    pass


def get_transaction_signers(hash_: UInt256) -> List[Signer]:
    """
    Gets the VM state of a transaction.

    :param hash_: hash identifier of the transaction
    :type hash_: UInt256
    :return: VM state of the transaction
    """
    pass


def get_transaction_vm_state(hash_: UInt256) -> VMState:
    """
    Gets the VM state of a transaction.

    :param hash_: hash identifier of the transaction
    :type hash_: UInt256
    :return: VM state of the transaction
    """
    pass


current_hash: UInt256 = UInt256()
"""
Gets the hash of the current block.

:meta hide-value:
"""

current_index: int = 0
"""
Gets the index of the current block.

:meta hide-value:
"""



from boa3.builtin.compile_time import public


@public
def main(ba1: bytearray, ba2: bytearray) -> bytearray:

    m = ba2[1:2]

    my_str = 'staoheustnau'

    m = my_str[3:5]

    m = ba1[1:len(ba1)]

    return m + my_str + ba2



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.neo import to_script_hash
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTemplate(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')
    OTHER_ACCOUNT_2 = to_script_hash(b'NNLi44dJNXtDNSBkofB48aTVYtb1zZrNEs')
    OTHER_ACCOUNT_3 = to_script_hash(b'NZcuGiwRu1QscpmCyxj5XwQBUf6sk7dJJN')

    def test_wrapped_neo_compile(self):
        path = self.get_contract_path('wrapped_neo.py')
        Boa3.compile(path)

    def test_wrapped_neo_symbol(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertEqual('zNEO', result)

    def test_wrapped_neo_decimals(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'decimals')
        self.assertEqual(8, result)

    def test_wrapped_neo_total_supply(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(total_supply, result)

    def test_wrapped_neo_total_balance_of(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(total_supply, result)

        # should fail when the script length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(10))
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(30))

    def test_wrapped_neo_total_transfer(self):
        transferred_amount = 10 * 10 ** 8  # 10 tokens
        path = self.get_contract_path('wrapped_neo.py')

        engine = TestEngine()

        # should fail if the sender doesn't sign
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount, "")
        self.assertEqual(False, result)
        wrapped_neo_address = engine.executed_script_hash.to_array()

        # should fail if the sender doesn't have enough balance
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, transferred_amount, "",
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        # should fail when any of the scripts' length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OWNER_SCRIPT_HASH, bytes(10), transferred_amount, "")
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    bytes(10), self.OTHER_ACCOUNT_1, transferred_amount, "")

        # should fail when the amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, -10, "")

        # fire the transfer event when transferring to yourself
        balance_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OWNER_SCRIPT_HASH, transferred_amount, "",
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer', origin=wrapped_neo_address)
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(self.OWNER_SCRIPT_HASH, receiver)
        self.assertEqual(transferred_amount, amount)

        # transferring to yourself doesn't change the balance
        balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(balance_before, balance_after)

        # does fire the transfer event when transferring to someone other than yourself
        balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_receiver_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount, "",
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(3, len(transfer_events))
        self.assertEqual(3, len(transfer_events[2].arguments))

        sender, receiver, amount = transfer_events[2].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(transferred_amount, amount)

        # transferring to someone other than yourself does change the balance
        balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_receiver_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(balance_sender_before - transferred_amount, balance_sender_after)
        self.assertEqual(balance_receiver_before + transferred_amount, balance_receiver_after)

    def test_wrapped_neo_verify(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        # should fail without signature
        result = self.run_smart_contract(engine, path, 'verify')
        self.assertEqual(False, result)

        # should fail if not signed by the owner
        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

    def test_wrapped_neo_burn(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        wrapped_neo_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_neo(wrapped_neo_address, 10_000_000 * 10 ** 8)
        burned_amount = 100 * 10 ** 8

        # burning zNEO will end up giving NEO to the one who burned it
        neo_wrapped_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', wrapped_neo_address)
        neo_owner_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', self.OWNER_SCRIPT_HASH)
        zneo_owner_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        # in this case, NEO will be given to the OWNER_SCRIPT_HASH
        result = self.run_smart_contract(engine, path, 'burn', self.OWNER_SCRIPT_HASH, burned_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertIsVoid(result)

        transfer_events = engine.get_events('Transfer', origin=wrapped_neo_address)
        self.assertGreaterEqual(len(transfer_events), 1)
        wrapped_token_transfer_event = transfer_events[-1]
        self.assertEqual(3, len(wrapped_token_transfer_event.arguments))

        sender, receiver, amount = wrapped_token_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(None, receiver)
        self.assertEqual(burned_amount, amount)

        transfer_events = engine.get_events('Transfer', origin=constants.NEO_SCRIPT)
        self.assertGreaterEqual(len(transfer_events), 1)
        neo_transfer_event = transfer_events[-1]
        self.assertEqual(3, len(neo_transfer_event.arguments))

        sender, receiver, amount = neo_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(wrapped_neo_address, sender)
        self.assertEqual(self.OWNER_SCRIPT_HASH, receiver)
        self.assertEqual(burned_amount, amount)

        # balance after burning
        neo_wrapped_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', wrapped_neo_address)
        neo_owner_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', self.OWNER_SCRIPT_HASH)
        zneo_owner_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(neo_wrapped_before - burned_amount, neo_wrapped_after)
        self.assertEqual(neo_owner_before + burned_amount, neo_owner_after)
        self.assertEqual(zneo_owner_before - burned_amount, zneo_owner_after)

        # should fail when the script length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'burn', bytes(15), burned_amount,
                                    signer_accounts=[self.OWNER_SCRIPT_HASH])
        # or amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'burn', self.OWNER_SCRIPT_HASH, -1,
                                    signer_accounts=[self.OWNER_SCRIPT_HASH])

    def test_wrapped_neo_approve(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        wrapped_neo_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        allowed_amount = 10 * 10 ** 8

        # this approve will fail, because OTHER_ACCOUNT_1 doesn't have enough zNEO
        result = self.run_smart_contract(engine, wrapped_neo_address, 'approve',
                                         self.OTHER_ACCOUNT_2, allowed_amount,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        # OWNER will give zNEO to OTHER_ACCOUNT_1 so that it can approve
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, allowed_amount, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # this approve will succeed, because OTHER_ACCOUNT_1 have enough zNEO
        result = self.run_smart_contract(engine, wrapped_neo_address, 'approve',
                                         self.OTHER_ACCOUNT_2, allowed_amount,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approved fired an event
        approval_events = engine.get_events('Approval')
        self.assertEqual(1, len(approval_events))

        owner, spender, amount = approval_events[0].arguments
        if isinstance(owner, str):
            owner = String(owner).to_bytes()
        if isinstance(spender, str):
            spender = String(spender).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, owner)
        self.assertEqual(self.OTHER_ACCOUNT_2, spender)
        self.assertEqual(allowed_amount, amount)

    def test_wrapped_neo_allowance(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        wrapped_neo_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        allowed_amount = 10 * 10 ** 8

        # OTHER_ACCOUNT_1 did not approve OTHER_SCRIPT_HASH
        result = self.run_smart_contract(engine, path, 'allowance', self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(0, result)

        # OWNER will give zNEO to OTHER_ACCOUNT_1 so that it can approve
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, allowed_amount, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # this approve will succeed, because OTHER_ACCOUNT_1 have enough zNEO
        result = self.run_smart_contract(engine, wrapped_neo_address, 'approve',
                                         self.OTHER_ACCOUNT_2, allowed_amount,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # OTHER_ACCOUNT_1 allowed OTHER_ACCOUNT_2 to spend transferred_amount of zNEO
        result = self.run_smart_contract(engine, path, 'allowance', self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(allowed_amount, result)

    def test_wrapped_neo_transfer_from(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        wrapped_neo_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        allowed_amount = 10 * 10 ** 8

        # OWNER will give zNEO to OTHER_ACCOUNT_3 so that it can approve another contracts
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_3, 10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(self.OTHER_ACCOUNT_3, receiver)
        self.assertEqual(10_000_000 * 100_000_000, amount)

        # this approve will succeed, because OTHER_ACCOUNT_3 have enough zNEO
        result = self.run_smart_contract(engine, wrapped_neo_address, 'approve',
                                         self.OTHER_ACCOUNT_1, allowed_amount,
                                         calling_script_hash=self.OTHER_ACCOUNT_3,
                                         signer_accounts=[self.OTHER_ACCOUNT_3])
        self.assertEqual(True, result)

        transferred_amount = allowed_amount

        # this transfer will fail,
        # because OTHER_ACCOUNT_1 is not allowed to transfer more than OTHER_ACCOUNT_3 approved
        result = self.run_smart_contract(engine, path, 'transferFrom', self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_3,
                                         self.OTHER_ACCOUNT_2, transferred_amount + 1 * 10 ** 8, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(2, len(transfer_events))

        # this transfer will succeed and will fire the Transfer event
        balance_spender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_3)
        balance_receiver_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_2)
        result = self.run_smart_contract(engine, path, 'transferFrom', self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_3,
                                         self.OTHER_ACCOUNT_2, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(3, len(transfer_events))
        self.assertEqual(3, len(transfer_events[2].arguments))

        sender, receiver, amount = transfer_events[2].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_3, sender)
        self.assertEqual(self.OTHER_ACCOUNT_2, receiver)
        self.assertEqual(transferred_amount, amount)

        # transferring changed the balance
        balance_spender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_3)
        balance_receiver_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_2)
        self.assertEqual(balance_spender_before, balance_spender_after)
        self.assertEqual(balance_sender_before - transferred_amount, balance_sender_after)
        self.assertEqual(balance_receiver_before + transferred_amount, balance_receiver_after)

        # OTHER_ACCOUNT_3 and OTHER_ACCOUNT_1 allowance was reduced to 0
        result = self.run_smart_contract(engine, path, 'allowance', self.OTHER_ACCOUNT_3, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_3])
        self.assertEqual(0, result)

        # this approve will succeed, because OTHER_ACCOUNT_3 have enough zNEO
        result = self.run_smart_contract(engine, wrapped_neo_address, 'approve',
                                         self.OTHER_ACCOUNT_1, allowed_amount,
                                         calling_script_hash=self.OTHER_ACCOUNT_3,
                                         signer_accounts=[self.OTHER_ACCOUNT_3])
        self.assertEqual(True, result)

        transferred_amount = allowed_amount - 4 * 10 ** 8

        result = self.run_smart_contract(engine, path, 'transferFrom', self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_3,
                                         self.OTHER_ACCOUNT_2, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # OTHER_ACCOUNT_3 and OTHER_ACCOUNT_1 allowance was reduced to allowed_amount - transferred_amount
        result = self.run_smart_contract(engine, path, 'allowance', self.OTHER_ACCOUNT_3, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_3])
        self.assertEqual(allowed_amount - transferred_amount, result)

        # should fail when any of the scripts' length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    self.OWNER_SCRIPT_HASH, bytes(10), self.OTHER_ACCOUNT_1, allowed_amount, None)
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    bytes(10), self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, allowed_amount, None)
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, bytes(10), allowed_amount, None)

        # should fail when the amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_2, -10, None)

    def test_wrapped_neo_on_nep17_payment(self):
        path = self.get_contract_path('wrapped_neo.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        wrapped_neo_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        minted_amount = 10 * 10 ** 8
        engine.add_neo(self.OTHER_ACCOUNT_1, minted_amount)

        # the smart contract will abort if some address other than NEO calls the onPayment method
        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, path, 'onNEP17Payment', self.OTHER_ACCOUNT_1, minted_amount, None,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        neo_wrapped_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', wrapped_neo_address)
        neo_aux_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', self.OTHER_ACCOUNT_1)
        zneo_aux_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        # transferring NEO to the wrapped_neo_address will mint them
        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, wrapped_neo_address, minted_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        transfer_events = engine.get_events('Transfer', origin=constants.NEO_SCRIPT)
        self.assertEqual(1, len(transfer_events))
        neo_transfer_event = transfer_events[0]
        self.assertEqual(3, len(neo_transfer_event.arguments))

        sender, receiver, amount = neo_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, sender)
        self.assertEqual(wrapped_neo_address, receiver)
        self.assertEqual(minted_amount, amount)

        transfer_events = engine.get_events('Transfer', origin=wrapped_neo_address)
        self.assertEqual(2, len(transfer_events))
        wrapped_token_transfer_event = transfer_events[1]
        self.assertEqual(3, len(wrapped_token_transfer_event.arguments))

        sender, receiver, amount = wrapped_token_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(None, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(minted_amount, amount)

        # balance after burning
        neo_wrapped_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', wrapped_neo_address)
        neo_aux_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', self.OTHER_ACCOUNT_1)
        zneo_aux_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(neo_wrapped_before + minted_amount, neo_wrapped_after)
        self.assertEqual(neo_aux_before - minted_amount, neo_aux_after)
        self.assertEqual(zneo_aux_before + minted_amount, zneo_aux_after)



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.model.builtin.interop.interopinterfacetype import InteropInterfaceType
from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.variable import Variable


class StorageContextType(InteropInterfaceType):
    """
    A class used to represent Neo StorageContext class
    """

    def __init__(self):
        super().__init__('StorageContext')

        self._variables: Dict[str, Variable] = {}
        self._instance_methods: Dict[str, Method] = {}
        self._constructor: Method = None

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        if len(self._instance_methods) == 0:
            from boa3.model.builtin.interop.storage.storagecontext.storagecontextcreatemapmethod import \
                StorageContextCreateMapMethod
            from boa3.model.builtin.interop.storage.storagecontext.storagecontextasreadonlymethod import \
                StorageContextAsReadOnlyMethod

            self._instance_methods = {
                'create_map': StorageContextCreateMapMethod(),
                'as_read_only': StorageContextAsReadOnlyMethod()
            }
        return self._instance_methods

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> StorageContextType:
        if value is None or cls._is_type_of(value):
            return _StorageContext

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, StorageContextType)


_StorageContext = StorageContextType()



from typing import List

from boa3.neo3.contracts.nef import MethodToken


class CompilerOutput:
    def __init__(self, bytecode: bytes, method_tokens: List[MethodToken] = None):
        self._bytecode = bytecode
        self._method_tokens = method_tokens if isinstance(method_tokens, list) else []

    @property
    def bytecode(self) -> bytes:
        return self._bytecode

    @property
    def method_tokens(self) -> List[MethodToken]:
        return self._method_tokens



"""
Use Upper Kebab Case to name the standards in this dictionary
"""
from boa3.model.standards.nep11standard import Nep11Standard
from boa3.model.standards.nep17standard import Nep17Standard

neo_standards = {
    'NEP-11': Nep11Standard(),
    'NEP-17': Nep17Standard()
}
